"""Simple Base for ASGI Apps."""

import inspect
import logging
import typing as t
from functools import partial

from http_router import Router as HTTPRouter, PrefixedRoute
from http_router.typing import TYPE_METHODS

from . import ASGIError, ASGINotFound, ASGIMethodNotAllowed, ASGIConnectionClosed, asgi_logger
from .middleware import (
    BaseMiddeware,
    LifespanMiddleware,
    StaticFilesMiddleware,
    parse_response
)
from .request import Request
from .response import ResponseError, Response
from .utils import to_awaitable, iscoroutinefunction
from .typing import Scope, Receive, Send, F


HTTP_METHODS = {'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH'}


class Router(HTTPRouter):
    """Rebind router errors."""

    NotFound: t.ClassVar[t.Type[Exception]] = ASGINotFound
    RouterError: t.ClassVar[t.Type[Exception]] = ASGIError
    MethodNotAllowed: t.ClassVar[t.Type[Exception]] = ASGIMethodNotAllowed


class HTTPView:
    """Class-based view pattern for handling HTTP method dispatching.

    .. code-block:: python

        @app.route('/custom')
        class CustomEndpoint(HTTPView):

            async def get(self, request):
                return 'Hello from GET'

            async def post(self, request):
                return 'Hello from POST'

        # ...
        async def test_my_endpoint(client):
            response = await client.get('/custom')
            assert await response.text() == 'Hello from GET'

            response = await client.post('/custom')
            assert await response.text() == 'Hello from POST'

            response = await client.put('/custom')
            assert response.status_code == 405

    """

    def __new__(cls, request: Request, *args, **opts):
        """Init the class and call it."""
        self = super().__new__(cls)
        return self(request, **opts)

    @classmethod
    def __route__(cls, router: Router, *paths: str,
                  methods: TYPE_METHODS = None, **params) -> t.Type['HTTPView']:
        """Bind the class view to the given router."""
        view_methods = dict(inspect.getmembers(cls, inspect.isfunction))
        methods = methods or [m for m in HTTP_METHODS if m.lower() in view_methods]
        return router.bind(cls, *paths, methods=methods, **params)

    def __call__(self, request: Request, *args, **opts) -> t.Awaitable:
        """Dispatch the given request by HTTP method."""
        method = getattr(self, request.method.lower())
        return method(request, **opts)


class AppInternalMiddleware(BaseMiddeware):
    """Process responses."""

    scopes = {'http'}

    async def __process__(self, scope: Scope, receive: Receive, send: Send):
        """Send ASGI messages."""
        response = await self.app(scope, receive, send)
        await response(scope, receive, send)


class App:
    """A helper to build ASGI Applications.

    Features:

    * Routing
    * ASGI-Tools :class:`Request`, :class:`Response`
    * Exception management
    * Static files
    * Lifespan events
    * Simplest middlewares

    :param debug: Enable debug mode (more logging, raise unhandled exceptions)
    :type debug: bool, False

    :param logger: Custom logger for the application
    :type logger: logging.Logger

    :param static_url_prefix: A prefix for static files
    :type static_url_prefix: str, "/static"

    :param static_folders: A list of folders to look static files
    :type static_folders: list[str]

    :param trim_last_slash: Consider "/path" and "/path/" as the same
    :type trim_last_slash: bool, False

    """

    exception_handlers: t.Dict[
        t.Union[int, t.Type[BaseException]],
        t.Callable[[Request, BaseException], t.Awaitable]
    ]

    def __init__(self, *, debug: bool = False,
                 logger: logging.Logger = asgi_logger,
                 static_url_prefix: str = '/static',
                 static_folders: t.Union[str, t.List[str]] = None, trim_last_slash: bool = False):
        """Initialize router and lifespan middleware."""

        # Register base exception handlers
        self.exception_handlers = {ASGIConnectionClosed: to_awaitable(lambda req, exc: None)}

        # Setup routing
        self.router = router = Router(
            trim_last_slash=trim_last_slash, validator=callable, converter=to_awaitable)

        # Setup logging
        self.logger = logger

        async def process(request: Request, receive: Receive, send: Send) -> t.Optional[Response]:
            """Find and call a callback, parse a response, handle exceptions."""
            scope = request.scope
            path = f"{ scope.get('root_path', '') }{ scope['path'] }"
            try:
                match = router(path, scope.get('method', 'GET'))
            except ASGINotFound:
                raise ResponseError.NOT_FOUND()
            except ASGIMethodNotAllowed:
                raise ResponseError.METHOD_NOT_ALLOWED()

            scope['path_params'] = {} if match.params is None else match.params
            response = await match.target(request)  # type: ignore
            if response is None and request['type'] == 'websocket':
                return None

            return parse_response(response)

        self.__internal__ = AppInternalMiddleware(process)  # type: ignore
        self.__process__ = self.__internal__.app

        # Setup lifespan
        self.lifespan = LifespanMiddleware(
            self.__internal__, ignore_errors=not debug, logger=self.logger)

        # Enable middleware for static files
        if static_folders and static_url_prefix:
            md = StaticFilesMiddleware.setup(folders=static_folders, url_prefix=static_url_prefix)
            self.middleware(md)

        # Debug mode
        self.debug = debug

        # Handle unknown exceptions
        if not debug:

            async def handle_unknown_exception(request: Request, exc: BaseException) -> Response:
                self.logger.exception(exc)
                return ResponseError.INTERNAL_SERVER_ERROR()

            self.exception_handlers[Exception] = handle_unknown_exception

        self.internal_middlewares: t.List = []

    def __route__(self, router: Router, *prefixes: str, methods: TYPE_METHODS = None, **params):
        """Mount self as a nested application."""
        def target(request):
            return self.__internal__(request, request.receive, request.send)

        for prefix in prefixes:
            route = RouteApp(prefix, set(), target=self)
            router.dynamic.insert(0, route)
        return self

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """Convert the given scope into a request and process."""
        scope['app'] = self
        request = Request(scope, receive, send)
        try:
            await self.lifespan(request, receive, send)
        except BaseException as exc:  # Handle exceptions
            response = await self.handle_exc(request, exc)
            if response is ...:
                raise

            await parse_response(response)(scope, receive, send)

    async def handle_exc(self, request: Request, exc: BaseException) -> t.Any:
        """Look for a handler for the given exception."""
        if isinstance(exc, Response) and exc.status_code in self.exception_handlers:
            return await self.exception_handlers[exc.status_code](request, exc)

        for etype in type(exc).mro():
            if etype in self.exception_handlers:
                return await self.exception_handlers[etype](request, exc)

        return exc if isinstance(exc, Response) else ...

    def middleware(self, md: F, insert_first: bool = False) -> F:
        """Register a middleware."""
        # Register as a simple middleware
        if iscoroutinefunction(md):

            if md not in self.internal_middlewares:
                if insert_first:
                    self.internal_middlewares.insert(0, md)
                else:
                    self.internal_middlewares.append(md)

            app = self.__process__
            for md in reversed(self.internal_middlewares):
                app = partial(md, app)

            self.__internal__.bind(app)

        else:
            self.lifespan.bind(md(self.lifespan.app))

        return md

    def route(self, *args, **kwargs) -> t.Callable:
        """Register a route."""
        return self.router.route(*args, **kwargs)

    def on_startup(self, fn: t.Callable):
        """Register a startup handler."""
        return self.lifespan.on_startup(fn)

    def on_shutdown(self, fn: t.Callable):
        """Register a shutdown handler."""
        return self.lifespan.on_shutdown(fn)

    def on_error(self, etype: t.Union[int, t.Type[BaseException]]):
        """Register an exception handler.

        .. code-block::

            @app.on_error(TimeoutError)
            async def timeout(request, error):
                return 'Something bad happens'

            @app.on_error(404)
            async def page_not_found(request, error):
                return render_template('page_not_found.html'), 404

        """
        def recorder(handler: F) -> F:
            self.exception_handlers[etype] = to_awaitable(handler)
            return handler

        return recorder


class RouteApp(PrefixedRoute):
    """Custom route to submount an application."""

    def __init__(self, path: str, methods: t.Set, target: App):
        """Create app callable."""
        path = path.rstrip('/')

        def app(request: Request):
            subrequest = request.__copy__(path=request.path[len(path):])
            receive = t.cast(Receive, request.receive)
            send = t.cast(Send, request.send)
            return target.__internal__.app(subrequest, receive, send)

        super(RouteApp, self).__init__(path, methods, app)



import binascii
import io
import mimetypes
import os
import typing as t
from collections import deque
from contextlib import asynccontextmanager
from functools import partial
from http import cookies
from json import loads, dumps
from pathlib import Path
from urllib.parse import urlencode

from yarl import URL

from . import ASGIConnectionClosed
from ._compat import aio_sleep, aio_spawn, aio_wait, aio_cancel, FIRST_COMPLETED
from .response import Response, ResponseWebSocket, parse_websocket_msg
from .utils import to_awaitable, parse_headers, CIMultiDict
from .typing import JSONType, Scope, Receive, Send, Message, ASGIApp


class TestResponse(Response):

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        self._receive = receive
        msg = await self._receive()
        assert msg.get('type') == 'http.response.start', 'Invalid Response'
        self.status_code = int(msg.get('status', 502))
        self.headers = parse_headers(msg.get('headers', []))  # type: ignore
        self.content_type = self.headers.get('content-type')
        for cookie in self.headers.getall('set-cookie', []):
            self.cookies.load(cookie)

    async def stream(self) -> t.AsyncGenerator[bytes, None]:
        """Stream the response."""
        more_body = True
        while more_body:
            msg = await self._receive()
            if msg.get('type') == 'http.response.body':
                chunk = msg.get('body')
                if chunk:
                    yield chunk
                more_body = msg.get('more_body', False)

    async def body(self) -> bytes:
        """Load response body."""
        _body = b''
        async for chunk in self.stream():
            _body += chunk
        return _body

    async def text(self) -> str:
        body = await self.body()
        return body.decode(self.charset)

    async def json(self) -> JSONType:
        text = await self.text()
        return loads(text)


class TestWebSocketResponse(ResponseWebSocket):

    # Disable app methods for clients
    accept = close = None  # type: ignore

    def connect(self) -> t.Coroutine[Message, t.Any, t.Any]:
        return self.send({'type': 'websocket.connect'})

    async def disconnect(self):
        await self.send({'type': 'websocket.disconnect', 'code': 1005})
        self.state = self.STATES.disconnected

    def send(self, msg, type='websocket.receive'):
        """Send a message to a client."""
        return super().send(msg, type=type)

    async def receive(self, raw=False):
        """Receive messages from a client."""
        if self.partner_state == self.STATES.disconnected:
            raise ASGIConnectionClosed

        msg = await self._receive()
        if not msg['type'].startswith('websocket.'):
            raise ValueError('Invalid websocket message: %r', msg)

        if msg['type'] == 'websocket.accept':
            self.partner_state = self.STATES.connected
            return await self.receive(raw=raw)

        if msg['type'] == 'websocket.close':
            self.partner_state == self.STATES.disconnected
            raise ASGIConnectionClosed('Connection has been closed.')

        return raw and msg or parse_websocket_msg(msg, charset=self.charset)


class ASGITestClient:
    """The test client allows you to make requests against an ASGI application.

    Features:

    * cookies
    * multipart/form-data
    * follow redirects
    * request streams
    * response streams
    * websocket support
    * lifespan management

    """

    def __init__(self, app: ASGIApp, base_url: str = 'http://localhost'):
        self.app = app
        self.base_url = URL(base_url)
        self.cookies: cookies.SimpleCookie = cookies.SimpleCookie()
        self.headers: t.Dict[str, str] = {}

    def __getattr__(self, name: str) -> t.Callable[..., t.Awaitable]:
        return partial(self.request, method=name.upper())

    async def request(
            self, path: str, method: str = 'GET', query: t.Union[str, t.Dict] = '',
            headers: t.Dict[str, str] = None, cookies: t.Dict = None,
            data: t.Union[bytes, str, t.Dict, t.AsyncGenerator] = b'', json: JSONType = None,
            follow_redirect: bool = True, timeout: float = 10.0) -> TestResponse:
        """Make a HTTP requests."""

        res = TestResponse()
        headers = headers or dict(self.headers)

        if isinstance(data, str):
            data = data.encode(res.charset)

        elif isinstance(data, dict):
            is_multipart = any(isinstance(value, io.IOBase) for value in data.values())
            if is_multipart:
                data, headers['Content-Type'] = encode_multipart(data)

            else:
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                data = urlencode(data).encode(res.charset)

        elif json is not None:
            headers['Content-Type'] = 'application/json'
            data = dumps(json).encode(res.charset)

        receive_from_client, send_to_app = simple_stream()
        receive_from_app, send_to_client = simple_stream()

        if isinstance(data, bytes):
            headers.setdefault('Content-Length', str(len(data)))

        scope = self.build_scope(
            path, headers=headers, query=query, cookies=cookies, type='http', method=method,
        )

        await aio_wait(
            aio_wait(
                self.app(scope, receive_from_client, send_to_client),
                stream_data(data, send_to_app, timeout),
            ), raise_timeout(timeout), strategy=FIRST_COMPLETED
        )
        await send_to_client({'type': 'http.response.body', 'more_body': False})
        await res(scope, receive_from_app, send_to_app)
        for n, v in res.cookies.items():
            self.cookies[n] = v

        if follow_redirect and res.status_code in {301, 302, 303, 307, 308}:
            return await self.get(res.headers['location'])

        return res

    # TODO: Timeouts for websockets
    @asynccontextmanager
    async def websocket(self, path: str, query: t.Union[str, t.Dict] = None,
                        headers: t.Dict = None, cookies: t.Dict = None):
        """Connect to a websocket."""
        receive_from_client, send_to_app = simple_stream()
        receive_from_app, send_to_client = simple_stream()

        ci_headers = CIMultiDict(headers or {})

        scope = self.build_scope(
            path, headers=ci_headers, query=query, cookies=cookies, type='websocket',
            subprotocols=ci_headers.get('Sec-WebSocket-Protocol', '').split(','),
        )
        ws = TestWebSocketResponse(scope, receive_from_app, send_to_app)
        async with aio_spawn(self.app, scope, receive_from_client, send_to_client):
            await ws.connect()
            yield ws
            await ws.disconnect()

    def lifespan(self, timeout: float = 3e-2):
        """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_ protocol.
        """
        return manage_lifespan(self.app, timeout=timeout)

    def build_scope(
            self, path: str, headers: t.Union[t.Dict, CIMultiDict] = None,
            query: t.Union[str, t.Dict] = None, cookies: t.Dict = None, **scope) -> Scope:
        """Prepare a request scope."""
        headers = headers or {}
        headers.setdefault('Remote-Addr', '127.0.0.1')
        headers.setdefault('User-Agent', 'ASGI-Tools-Test-Client')
        headers.setdefault('Host', self.base_url.host)

        if cookies:
            for c, v in cookies.items():
                self.cookies[c] = v

        if len(self.cookies):
            headers.setdefault('Cookie', self.cookies.output(header='', sep=';'))

        url = URL(path)
        if query:
            url = url.with_query(query)

        return dict({
            'asgi': {'version': '3.0'},
            'http_version': '1.1',
            'path': url.path,
            'query_string': url.query_string.encode('latin-1'),
            'raw_path': url.raw_path.encode('latin-1'),
            'root_path': '',
            'scheme': scope.get('type') == 'http' and self.base_url.scheme or 'ws',
            'headers': [
                (key.lower().encode('latin-1'), str(val).encode('latin-1'))
                for key, val in (headers or {}).items()
            ],
            'server': ('127.0.0.1', self.base_url.port),
        }, **scope)


def encode_multipart(data: t.Dict) -> t.Tuple[bytes, str]:
    body = io.BytesIO()
    boundary = binascii.hexlify(os.urandom(16))
    for name, value in data.items():
        headers = f'Content-Disposition: form-data; name="{ name }"'
        if hasattr(value, 'read'):
            filename = getattr(value, 'name', None)
            if filename:
                headers = f'{ headers }; filename="{ Path(filename).name }"'
                content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
                headers = f'{ headers }\r\nContent-Type: { content_type }'
            value = value.read()

        body.write(b"--%b\r\n" % boundary)
        body.write(headers.encode('utf-8'))
        body.write(b'\r\n\r\n')
        if isinstance(value, str):
            value = value.encode('utf-8')
        body.write(value)
        body.write(b"\r\n")

    body.write(b"--%b--\r\n" % boundary)
    return body.getvalue(), (b"multipart/form-data; boundary=%s" % boundary).decode()


def simple_stream(maxlen=None):
    queue = deque(maxlen=maxlen)

    async def receive():
        while not queue:
            await aio_sleep(1e-3)
        return queue.popleft()

    return receive, to_awaitable(queue.append)


async def raise_timeout(timeout: t.Union[int, float]):
    await aio_sleep(timeout)
    raise TimeoutError('Timeout occured')


async def stream_data(data: t.Union[bytes, t.AsyncGenerator[t.Any, bytes]],
                      send: t.Callable[..., t.Awaitable], timeout: t.Union[int, float]):
    """Stream a data to an application."""

    if isinstance(data, bytes):
        return await send({'type': 'http.request', 'body': data, 'more_body': False})

    async for chunk in data:
        await send({'type': 'http.request', 'body': chunk, 'more_body': True})
    await send({'type': 'http.request', 'body': b'', 'more_body': False})


@asynccontextmanager
async def manage_lifespan(app, timeout: float = 3e-2):
    """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_ protocol.
    """
    receive_from_client, send_to_app = simple_stream()
    receive_from_app, send_to_client = simple_stream()

    async def safe_spawn():
        try:
            await app({'type': 'lifespan'}, receive_from_client, send_to_client)
        except BaseException:
            pass

    async with aio_spawn(safe_spawn) as task:
        await send_to_app({'type': 'lifespan.startup'})
        msg = await aio_wait(
            receive_from_app(), aio_sleep(timeout), strategy=FIRST_COMPLETED)
        if msg and isinstance(msg, t.Mapping):
            if msg['type'] == 'lifespan.startup.failed':
                await aio_cancel(task)
            else:
                assert msg['type'] == 'lifespan.startup.complete'

        yield

        await send_to_app({'type': 'lifespan.shutdown'})
        msg = await aio_wait(receive_from_app(), aio_sleep(timeout), strategy=FIRST_COMPLETED)
        if msg and isinstance(msg, t.Mapping):
            assert msg['type'] == 'lifespan.shutdown.complete'


# pylama:ignore=D



""" ASGI-Tools -- Tools to make ASGI Applications """

__version__ = "0.62.1"
__license__ = "MIT"

import logging


asgi_logger: logging.Logger = logging.getLogger('asgi-tools')


class ASGIError(Exception):
    """Base class for ASGI-Tools Errors."""

    pass


class ASGIConnectionClosed(ASGIError):
    """ASGI-Tools connection closed error."""

    pass


class ASGIDecodeError(ASGIError, ValueError):
    """ASGI-Tools decoding error."""

    pass


class ASGINotFound(ASGIError):
    """Raise when http handler not found."""

    pass


class ASGIMethodNotAllowed(ASGIError):
    """Raise when http method not found."""

    pass


DEFAULT_CHARSET: str = 'utf-8'

from .request import Request  # noqa
from .response import (  # noqa
    Response, ResponseHTML, ResponseJSON, ResponseText, ResponseRedirect, ResponseError,
        ResponseStream, ResponseSSE, ResponseFile, ResponseWebSocket, parse_response
)
from .middleware import (  # noqa
    RequestMiddleware, ResponseMiddleware, LifespanMiddleware,
    RouterMiddleware, StaticFilesMiddleware
)
from .app import App, HTTPView  # noqa

from http_router import NotFound, MethodNotAllowed  # noqa



"""ASGI responses."""

from __future__ import annotations

from email.utils import formatdate
from enum import Enum
from functools import partial
from hashlib import md5
from http import cookies, HTTPStatus
from mimetypes import guess_type
from multidict import MultiDict
from pathlib import Path
from urllib.parse import quote, quote_plus
from stat import S_ISDIR
import os
import typing as t

from . import DEFAULT_CHARSET, ASGIError, ASGIConnectionClosed
from ._compat import aio_wait, FIRST_COMPLETED, aio_stream_file, json_dumps
from .request import Request
from .typing import Message, ResponseContent, Scope, Receive, Send


class Response:
    """A base class to make ASGI_ responses.

    :param content: A response's body
    :type content: str | bytes
    :param status_code: An HTTP status code
    :type status_code: int
    :param headers: A dictionary of HTTP headers
    :type headers: dict[str, str]
    :param content_type: A string with the content-type
    :type content_type: str
    """

    headers: MultiDict  #: Multidict of response's headers
    cookies: cookies.SimpleCookie
    """ Set/Update cookies

    * `response.cookies[name] = value` ``str`` -- set a cookie's value
    * `response.cookies[name]['path'] = value` ``str`` -- set a cookie's path
    * `response.cookies[name]['expires'] = value` ``int`` -- set a cookie's expire
    * `response.cookies[name]['domain'] = value` ``str`` -- set a cookie's domain
    * `response.cookies[name]['max-age'] = value` ``int`` -- set a cookie's max-age
    * `response.cookies[name]['secure'] = value` ``bool``-- is the cookie should only be sent if request is SSL
    * `response.cookies[name]['httponly'] = value` ``bool`` -- is the cookie should be available through HTTP request only (not from JS)
    * `response.cookies[name]['samesite'] = value` ``str`` -- set a cookie's strategy ('lax'|'strict'|'none')

    """
    charset: str = DEFAULT_CHARSET
    content_type: t.Optional[str] = None
    status_code: int = HTTPStatus.OK.value

    def __init__(
            self, content: ResponseContent = None, status_code: int = None,
            headers: dict = None, content_type: str = None):
        """Setup the response."""
        self.content = content
        self.headers: MultiDict = MultiDict(headers or {})
        self.cookies: cookies.SimpleCookie = cookies.SimpleCookie()
        if status_code is not None:
            self.status_code = status_code

        content_type = content_type or self.content_type
        if content_type:
            self.headers.setdefault(
                'content-type', content_type.startswith('text/') and
                f"{content_type}; charset={self.charset}" or content_type
            )

    def __str__(self) -> str:
        """Stringify the response."""
        return f"{self.status_code}"

    def __repr__(self) -> str:
        """Stringify the response."""
        return f"<{ self.__class__.__name__ } '{ self }'>"

    async def __call__(self, scope: t.Any, receive: t.Any, send: Send) -> None:
        """Behave as an ASGI application."""
        self.headers.setdefault('content-length', str(len(self.__content__)))

        await send(self.msg_start())
        await send({"type": "http.response.body", "body": self.__content__})

    @property
    def content(self):
        """Get self content."""
        return self.__content__

    @content.setter
    def content(self, content: ResponseContent):
        # py38 return :=
        if isinstance(content, str):
            self.__content__ = content.encode(self.charset)

        elif content is None:
            self.__content__ = b""

        elif isinstance(content, bytes):
            self.__content__ = content

        else:
            self.__content__ = str(content).encode(self.charset)

    def msg_start(self) -> Message:
        """Get ASGI response start message."""
        headers = [
            (key.encode('latin-1'), str(val).encode('latin-1'))
            for key, val in self.headers.items()
        ]

        for cookie in self.cookies.values():
            headers.append((b"set-cookie", cookie.output(header='').strip().encode('latin-1')))

        return {
            "type": "http.response.start",
            "status": self.status_code,
            "headers": headers,
        }


class ResponseText(Response):
    """A helper to return plain text responses (text/plain)."""

    content_type = 'text/plain'


class ResponseHTML(Response):
    """A helper to return HTML responses (text/html)."""

    content_type = 'text/html'


class ResponseJSON(Response):
    """A helper to return JSON responses (application/json).

    The class optionally supports `ujson <https://pypi.org/project/ujson/>`_ and `orjson
    <https://pypi.org/project/orjson/>`_ JSON libraries. Install one of them to use instead
    the standard library.

    """

    content_type = 'application/json'

    def dumps(self, content: ResponseContent) -> bytes:
        """Dumps the given content."""
        return json_dumps(content)

    @Response.content.setter  # type: ignore
    def content(self, content: ResponseContent):
        """Jsonify the content."""
        self.__content__ = self.dumps(content)


class ResponseStream(Response):
    """A helper to stream a response's body.

    :param content: An async generator to stream the response's body
    :type content: AsyncGenerator
    """

    @Response.content.setter  # type: ignore
    def content(self, content: t.AsyncGenerator[ResponseContent, None] = None):
        """Store self content as is."""
        self.__content__ = content  # type: ignore

    def prepare_chunk(self, chunk: t.Any) -> bytes:
        """Prepare a chunk from stream generator to send."""
        if not isinstance(chunk, bytes):
            chunk = str(chunk).encode(self.charset)
        return chunk

    async def listen_for_disconnect(self, receive: Receive):
        """Listen for the client has been disconnected."""
        while True:
            message = await receive()
            if message['type'] == 'http.disconnect':
                break

    async def stream_response(self, send: Send):
        """Stream response content."""
        await send(self.msg_start())
        if self.content:
            async for chunk in self.content:
                await send({"type": "http.response.body",
                            "body": self.prepare_chunk(chunk), "more_body": True})

        await send({"type": "http.response.body", "body": b""})

    async def __call__(self, scope: t.Any, receive: t.Any, send: Send) -> None:
        """Behave as an ASGI application."""
        await aio_wait(
            self.listen_for_disconnect(receive),
            self.stream_response(send),
            strategy=FIRST_COMPLETED,
        )


class ResponseSSE(ResponseStream):
    """A helper to stream SSE (server side events).

    :param content: An async generator to stream the events
    :type content: AsyncGenerator
    """

    content_type = 'text/event-stream'

    def msg_start(self) -> Message:
        """Set cache-control header."""
        self.headers.setdefault('Cache-Control', 'no-cache')
        return super(ResponseSSE, self).msg_start()

    def prepare_chunk(self, chunk: t.Any) -> bytes:
        """Prepare a chunk from stream generator to send."""
        if isinstance(chunk, dict):
            chunk = '\n'.join(f"{k}: {v}" for k, v in chunk.items())

        if not isinstance(chunk, bytes):
            chunk = str(chunk).encode(self.charset)

        return chunk + b'\n\n'


class ResponseFile(ResponseStream):
    """A helper to stream files as a response body.

    :param filepath: The filepath to the file
    :type filepath: str | Path
    :param chunk_size: Default chunk size (32768)
    :type chunk_size: int
    :param filename: If set, `Content-Disposition` header will be generated
    :type filename: str
    :param headers_only: Return only file headers
    :type headers_only: bool

    """

    def __init__(self, filepath: t.Union[str, Path], *, chunk_size: int = 32 * 1024,
                 filename: str = None, headers_only: bool = False, **kwargs) -> None:
        """Store filepath to self."""
        try:
            stat = os.stat(filepath)
        except FileNotFoundError as exc:
            raise ASGIError(*exc.args)

        if S_ISDIR(stat.st_mode):
            raise ASGIError(f"It's a directory: {filepath}")

        stream = aio_stream_file(filepath, chunk_size) if not headers_only else None
        super(ResponseFile, self).__init__(stream, **kwargs)  # type: ignore

        headers = self.headers
        if filename and 'content-disposition' not in headers:
            headers['content-disposition'] = f'attachment; filename="{quote(filename)}"'

        if 'content-type' not in headers:
            headers['content-type'] = guess_type(filename or str(filepath))[0] or "text/plain"

        headers.setdefault('content-length', str(stat.st_size))
        headers.setdefault('last-modified', formatdate(stat.st_mtime, usegmt=True))
        etag = str(stat.st_mtime) + "-" + str(stat.st_size)
        headers.setdefault('etag', md5(etag.encode()).hexdigest())


class ResponseWebSocket(Response):
    """A helper to work with websockets.

    :param scope: Request info (ASGI Scope | ASGI-Tools Request)
    :type scope: dict
    :param receive: ASGI receive function
    :param send: ASGI send function
    """

    class STATES(Enum):
        """Represent websocket states."""

        connecting = 0
        connected = 1
        disconnected = 2

    def __init__(self, scope: Scope, receive: Receive = None, send: Send = None) -> None:
        """Initialize the websocket response."""
        if isinstance(scope, Request):
            receive, send = scope.receive, scope.send

        super(ResponseWebSocket, self).__init__()
        assert receive and send, 'Invalid initialization'
        self._receive: Receive = receive
        self._send: Send = send
        self.state = self.STATES.connecting
        self.partner_state = self.STATES.connecting

    async def __call__(self, scope: t.Any, receive: t.Any, send: Send):
        """Close websocket if the response has been returned."""
        await send({'type': 'websocket.close'})

    async def __aenter__(self):
        """Use it as async context manager."""
        await self.accept()
        return self

    async def __aexit__(self, *args):
        """Use it as async context manager."""
        await self.close()

    @property
    def connected(self) -> bool:
        """Check that is the websocket connected."""
        return self.state == self.STATES.connected and self.partner_state == self.STATES.connected

    async def _connect(self) -> bool:
        """Wait for connect message."""
        if self.partner_state == self.STATES.connecting:
            msg = await self._receive()
            assert msg.get('type') == 'websocket.connect'
            self.partner_state = self.STATES.connected

        return self.partner_state == self.STATES.connected

    async def accept(self, **params) -> None:
        """Accept a websocket connection."""
        if self.partner_state == self.STATES.connecting:
            await self._connect()

        await self.send({'type': 'websocket.accept', **params})
        self.state = self.STATES.connected

    async def close(self, code: int = 1000) -> None:
        """Sent by the application to tell the server to close the connection."""
        if self.connected:
            await self.send({'type': 'websocket.close', 'code': code})
        self.state = self.STATES.disconnected

    async def send(self, msg: t.Union[t.Dict, str, bytes], type='websocket.send') -> None:
        """Send the given message to a client."""
        if self.state == self.STATES.disconnected:
            raise ASGIConnectionClosed('Cannot send once the connection has been disconnected.')

        if not isinstance(msg, dict):
            msg = {'type': type, (isinstance(msg, str) and 'text' or 'bytes'): msg}

        return await self._send(msg)

    async def send_json(self, data) -> None:
        """Serialize the given data to JSON and send to a client."""
        return await self._send({'type': 'websocket.send', 'bytes': json_dumps(data)})

    async def receive(self, raw: bool = False) -> t.Union[Message, str]:
        """Receive messages from a client.

        :param raw: Receive messages as is.
        """
        if self.partner_state == self.STATES.disconnected:
            raise ASGIConnectionClosed('Cannot receive once a connection has been disconnected.')

        if self.partner_state == self.STATES.connecting:
            await self._connect()
            return await self.receive(raw=raw)

        msg = await self._receive()
        if msg['type'] == 'websocket.disconnect':
            self.partner_state = self.STATES.disconnected

        return raw and msg or parse_websocket_msg(msg, charset=self.charset)


class ResponseRedirect(Response, BaseException):
    """A helper to return HTTP redirects. Uses a 307 status code by default.

    :param url: A string with the new location
    :type url: str
    """

    status_code: int = HTTPStatus.TEMPORARY_REDIRECT.value

    def __init__(self, url: str, status_code: int = None, **kwargs) -> None:
        """Set status code and prepare location."""
        super(ResponseRedirect, self).__init__(status_code=status_code, **kwargs)
        assert 300 <= self.status_code < 400, f"Invalid status code for redirection: {self.status_code}"  # noqa
        self.headers["location"] = quote_plus(str(url), safe=":/%#?&=@[]!$&'()*+,;")


class ResponseErrorMeta(type):
    """Generate Response Errors by HTTP names."""

    # XXX: From python 3.9 -> partial['ResponseError]
    def __getattr__(cls, name: str) -> t.Callable[..., ResponseError]:
        """Generate Response Errors by HTTP names."""
        status = HTTPStatus[name]
        return partial(cls, status_code=status.value)


class ResponseError(Response, BaseException, metaclass=ResponseErrorMeta):
    """A helper to return HTTP errors. Uses a 500 status code by default.

    :param message: A string with the error's message (HTTPStatus messages will be used by default)
    :type message: str

    You able to use :py:class:`http.HTTPStatus` properties with the `ResponseError` class

    .. code-block:: python

        response = ResponseError.BAD_REQUEST('invalid data')
        response = ResponseError.NOT_FOUND()
        response = ResponseError.BAD_GATEWAY()
        # and etc

    """

    status_code: int = HTTPStatus.INTERNAL_SERVER_ERROR.value

    # Typing annotations
    if t.TYPE_CHECKING:
        BAD_REQUEST: t.Callable[..., ResponseError]                       # 400
        UNAUTHORIZED: t.Callable[..., ResponseError]                      # 401
        PAYMENT_REQUIRED: t.Callable[..., ResponseError]                  # 402
        FORBIDDEN: t.Callable[..., ResponseError]                         # 403
        NOT_FOUND: t.Callable[..., ResponseError]                         # 404
        METHOD_NOT_ALLOWED: t.Callable[..., ResponseError]                # 405
        NOT_ACCEPTABLE: t.Callable[..., ResponseError]                    # 406
        PROXY_AUTHENTICATION_REQUIRED: t.Callable[..., ResponseError]     # 407
        REQUEST_TIMEOUT: t.Callable[..., ResponseError]                   # 408
        CONFLICT: t.Callable[..., ResponseError]                          # 409
        GONE: t.Callable[..., ResponseError]                              # 410
        LENGTH_REQUIRED: t.Callable[..., ResponseError]                   # 411
        PRECONDITION_FAILED: t.Callable[..., ResponseError]               # 412
        REQUEST_ENTITY_TOO_LARGE: t.Callable[..., ResponseError]          # 413
        REQUEST_URI_TOO_LONG: t.Callable[..., ResponseError]              # 414
        UNSUPPORTED_MEDIA_TYPE: t.Callable[..., ResponseError]            # 415
        REQUESTED_RANGE_NOT_SATISFIABLE: t.Callable[..., ResponseError]   # 416
        EXPECTATION_FAILED: t.Callable[..., ResponseError]                # 417
        # XXX: From python 3.9
        # IM_A_TEAPOT: t.Callable[..., ResponseError]                       # 418
        # MISDIRECTED_REQUEST: t.Callable[..., ResponseError]               # 421
        UNPROCESSABLE_ENTITY: t.Callable[..., ResponseError]              # 422
        LOCKED: t.Callable[..., ResponseError]                            # 423
        FAILED_DEPENDENCY: t.Callable[..., ResponseError]                 # 424
        TOO_EARLY: t.Callable[..., ResponseError]                         # 425
        UPGRADE_REQUIRED: t.Callable[..., ResponseError]                  # 426
        PRECONDITION_REQUIRED: t.Callable[..., ResponseError]             # 428
        TOO_MANY_REQUESTS: t.Callable[..., ResponseError]                 # 429
        REQUEST_HEADER_FIELDS_TOO_LARGE: t.Callable[..., ResponseError]   # 431
        # XXX: From python 3.9
        # UNAVAILABLE_FOR_LEGAL_REASONS: t.Callable[..., ResponseError]     # 451

        INTERNAL_SERVER_ERROR: t.Callable[..., ResponseError]             # 500
        NOT_IMPLEMENTED: t.Callable[..., ResponseError]                   # 501
        BAD_GATEWAY: t.Callable[..., ResponseError]                       # 502
        SERVICE_UNAVAILABLE: t.Callable[..., ResponseError]               # 503
        GATEWAY_TIMEOUT: t.Callable[..., ResponseError]                   # 504
        HTTP_VERSION_NOT_SUPPORTED: t.Callable[..., ResponseError]        # 505
        VARIANT_ALSO_NEGOTIATES: t.Callable[..., ResponseError]           # 506
        INSUFFICIENT_STORAGE: t.Callable[..., ResponseError]              # 507
        LOOP_DETECTED: t.Callable[..., ResponseError]                     # 508
        NOT_EXTENDED: t.Callable[..., ResponseError]                      # 510
        NETWORK_AUTHENTICATION_REQUIRED: t.Callable[..., ResponseError]   # 511

    def __init__(self, message: ResponseContent = None, status_code: int = None, **kwargs):
        """Check error status."""
        super(ResponseError, self).__init__(content=message, status_code=status_code, **kwargs)
        assert self.status_code >= 400, f"Invalid status code for an error: {self.status_code}"
        self.content = message or HTTPStatus(self.status_code).description


CAST_RESPONSE: t.Dict[t.Type, t.Type[Response]] = {
    bool: ResponseJSON,
    bytes: ResponseHTML,
    dict: ResponseJSON,
    int: ResponseJSON,
    list: ResponseJSON,
    str: ResponseHTML,
    type(None): ResponseJSON,
}


def parse_response(response: t.Any, headers: t.Dict = None) -> Response:
    """Parse the given object and convert it into a asgi_tools.Response."""
    rtype = type(response)
    if issubclass(rtype, Response):
        return response

    ResponseType = CAST_RESPONSE.get(rtype)
    if ResponseType:
        return ResponseType(response, headers=headers)

    if rtype is tuple:
        status, *contents = response
        assert isinstance(status, int), 'Invalid Response Status'
        if len(contents) > 1:
            headers, *contents = contents
        response = parse_response(contents[0] or '' if contents else '', headers=headers)
        response.status_code = status
        return response

    return ResponseText(str(response), headers=headers)


def parse_websocket_msg(msg: Message, charset: str = None) -> t.Union[Message, str]:
    """Prepare websocket message."""
    data = msg.get('text')
    if data:
        return data

    data = msg.get('bytes')
    if data:
        return data.decode(charset)

    return msg

# pylama: ignore=E501
